@@ -1,7 +1,7 @@
 use {
     crate::InstructionReader,
     koto_parser::{ConstantPool, Span},
-    std::{path::PathBuf, sync::Arc},
+    std::{path::PathBuf, rc::Rc},
 };
 
 /// Debug information for a Koto program
@@ -13,6 +13,10 @@ pub struct DebugInfo {
 }
 
 impl DebugInfo {
+    /// Adds a span to the source map for a given ip
+    ///
+    /// Instructions with matching spans share the same entry, so if the span matches the
+    /// previously pushed span then this is a no-op.
     pub fn push(&mut self, ip: usize, span: Span) {
         if let Some(entry) = self.source_map.last() {
             if entry.1 == span {
@@ -49,8 +53,6 @@ pub struct Chunk {
     pub bytes: Vec<u8>,
     /// The constant data associated with the chunk's bytecode
     pub constants: ConstantPool,
-    /// The constant string data associated with the chunk's bytecode
-    pub string_constants_arc: Arc<str>,
     /// The path of the program's source file
     pub source_path: Option<PathBuf>,
     /// Debug information associated with the chunk's bytecode
@@ -62,7 +64,6 @@ impl Default for Chunk {
         Self {
             bytes: vec![],
             constants: ConstantPool::default(),
-            string_constants_arc: String::default().into(),
             source_path: None,
             debug_info: DebugInfo::default(),
         }
@@ -70,6 +71,7 @@ impl Default for Chunk {
 }
 
 impl Chunk {
+    /// Initializes a Chunk
     pub fn new(
         bytes: Vec<u8>,
         constants: ConstantPool,
@@ -78,7 +80,6 @@ impl Chunk {
     ) -> Self {
         Self {
             bytes,
-            string_constants_arc: constants.string_data().into(),
             constants,
             source_path,
             debug_info,
@@ -86,7 +87,7 @@ impl Chunk {
     }
 
     /// Returns a [String] displaying the instructions contained in the compiled [Chunk]
-    pub fn bytes_as_string(chunk: Arc<Chunk>) -> String {
+    pub fn bytes_as_string(chunk: Rc<Chunk>) -> String {
         let mut iter = chunk.bytes.iter();
         let mut result = String::new();
 
@@ -111,7 +112,7 @@ impl Chunk {
     }
 
     /// Returns a [String] displaying the annotated instructions contained in the compiled [Chunk]
-    pub fn instructions_as_string(chunk: Arc<Chunk>, source_lines: &[&str]) -> String {
+    pub fn instructions_as_string(chunk: Rc<Chunk>, source_lines: &[&str]) -> String {
         let mut result = String::new();
         let mut reader = InstructionReader::new(chunk);
         let mut ip = reader.ip;
