@@ -1,11 +1,12 @@
 use {
     crate::{Chunk, Op},
     koto_parser::{ConstantIndex, MetaKeyId},
-    std::{convert::TryInto, fmt, sync::Arc},
+    std::{convert::TryInto, fmt, rc::Rc},
 };
 
 #[derive(Debug)]
 #[repr(u8)]
+#[allow(missing_docs)]
 pub enum TypeId {
     List,
     Tuple,
@@ -23,25 +24,39 @@ impl TypeId {
     }
 }
 
+/// Flags used to define the properties of a Function
 pub struct FunctionFlags {
+    /// True if the function is an instance function
     pub instance_function: bool,
+    /// True if the function has a variadic argument
     pub variadic: bool,
+    /// True if the function is a generator
     pub generator: bool,
+    /// True if the function has a single argument which is an unpacked tuple
+    pub arg_is_unpacked_tuple: bool,
 }
 
 impl FunctionFlags {
-    pub const INSTANCE: u8 = 0b0000001;
-    pub const VARIADIC: u8 = 0b0000010;
-    pub const GENERATOR: u8 = 0b0000100;
+    /// Corresponding to [FunctionFlags::instance_function]
+    pub const INSTANCE: u8 = 1 << 0;
+    /// Corresponding to [FunctionFlags::variadic]
+    pub const VARIADIC: u8 = 1 << 1;
+    /// Corresponding to [FunctionFlags::generator]
+    pub const GENERATOR: u8 = 1 << 2;
+    /// Corresponding to [FunctionFlags::arg_is_unpacked_tuple]
+    pub const ARG_IS_UNPACKED_TUPLE: u8 = 1 << 3;
 
+    /// Initializes a flags struct from a byte
     pub fn from_byte(byte: u8) -> Self {
         Self {
-            instance_function: byte & Self::INSTANCE == Self::INSTANCE,
-            variadic: byte & Self::VARIADIC == Self::VARIADIC,
-            generator: byte & Self::GENERATOR == Self::GENERATOR,
+            instance_function: byte & Self::INSTANCE != 0,
+            variadic: byte & Self::VARIADIC != 0,
+            generator: byte & Self::GENERATOR != 0,
+            arg_is_unpacked_tuple: byte & Self::ARG_IS_UNPACKED_TUPLE != 0,
         }
     }
 
+    /// Returns a byte containing the packed flags
     pub fn as_byte(&self) -> u8 {
         let mut result = 0;
         if self.instance_function {
@@ -53,11 +68,17 @@ impl FunctionFlags {
         if self.generator {
             result |= Self::GENERATOR;
         }
+        if self.arg_is_unpacked_tuple {
+            result |= Self::ARG_IS_UNPACKED_TUPLE;
+        }
         result
     }
 }
 
 /// Decoded instructions produced by an [InstructionReader] for execution in the runtime
+///
+/// For descriptions of each instruction's purpose, see corresponding [Op] entries.
+#[allow(missing_docs)]
 pub enum Instruction {
     Error {
         message: String,
@@ -111,13 +132,13 @@ pub enum Instruction {
     },
     MakeNum2 {
         register: u8,
-        count: u8,
         element_register: u8,
+        count: u8,
     },
     MakeNum4 {
         register: u8,
-        count: u8,
         element_register: u8,
+        count: u8,
     },
     SequenceStart {
         register: u8,
@@ -179,6 +200,7 @@ pub enum Instruction {
         instance_function: bool,
         variadic: bool,
         generator: bool,
+        arg_is_unpacked_tuple: bool,
         size: usize,
     },
     Capture {
@@ -188,7 +210,11 @@ pub enum Instruction {
     },
     Negate {
         register: u8,
-        source: u8,
+        value: u8,
+    },
+    Not {
+        register: u8,
+        value: u8,
     },
     Add {
         register: u8,
@@ -256,23 +282,18 @@ pub enum Instruction {
     JumpBack {
         offset: usize,
     },
-    JumpBackIf {
-        register: u8,
-        offset: usize,
-        jump_condition: bool,
-    },
     Call {
         result: u8,
         function: u8,
         frame_base: u8,
         arg_count: u8,
     },
-    CallChild {
+    CallInstance {
         result: u8,
         function: u8,
         frame_base: u8,
         arg_count: u8,
-        parent: u8,
+        instance: u8,
     },
     Return {
         register: u8,
@@ -301,7 +322,7 @@ pub enum Instruction {
         iterator: u8,
         jump_offset: usize,
     },
-    ValueIndex {
+    TempIndex {
         register: u8,
         value: u8,
         index: i8,
@@ -360,6 +381,11 @@ pub enum Instruction {
         value: u8,
     },
     Access {
+        register: u8,
+        value: u8,
+        key: ConstantIndex,
+    },
+    AccessString {
         register: u8,
         value: u8,
         key: u8,
@@ -383,6 +409,7 @@ pub enum Instruction {
     },
     StringStart {
         register: u8,
+        size_hint: usize,
     },
     StringPush {
         register: u8,
@@ -431,6 +458,7 @@ impl fmt::Display for Instruction {
             Function { .. } => write!(f, "Function"),
             Capture { .. } => write!(f, "Capture"),
             Negate { .. } => write!(f, "Negate"),
+            Not { .. } => write!(f, "Not"),
             Add { .. } => write!(f, "Add"),
             Subtract { .. } => write!(f, "Subtract"),
             Multiply { .. } => write!(f, "Multiply"),
@@ -445,9 +473,8 @@ impl fmt::Display for Instruction {
             Jump { .. } => write!(f, "Jump"),
             JumpIf { .. } => write!(f, "JumpIf"),
             JumpBack { .. } => write!(f, "JumpBack"),
-            JumpBackIf { .. } => write!(f, "JumpBackIf"),
             Call { .. } => write!(f, "Call"),
-            CallChild { .. } => write!(f, "CallChild"),
+            CallInstance { .. } => write!(f, "CallInstance"),
             Return { .. } => write!(f, "Return"),
             Yield { .. } => write!(f, "Yield"),
             Throw { .. } => write!(f, "Throw"),
@@ -455,7 +482,7 @@ impl fmt::Display for Instruction {
             IterNext { .. } => write!(f, "IterNext"),
             IterNextTemp { .. } => write!(f, "IterNextTemp"),
             IterNextQuiet { .. } => write!(f, "IterNextQuiet"),
-            ValueIndex { .. } => write!(f, "ValueIndex"),
+            TempIndex { .. } => write!(f, "TempIndex"),
             SliceFrom { .. } => write!(f, "SliceFrom"),
             SliceTo { .. } => write!(f, "SliceTo"),
             IsTuple { .. } => write!(f, "IsTuple"),
@@ -468,6 +495,7 @@ impl fmt::Display for Instruction {
             MetaExport { .. } => write!(f, "MetaExport"),
             MetaExportNamed { .. } => write!(f, "MetaExportNamed"),
             Access { .. } => write!(f, "Access"),
+            AccessString { .. } => write!(f, "AccessString"),
             TryStart { .. } => write!(f, "TryStart"),
             TryEnd => write!(f, "TryEnd"),
             Debug { .. } => write!(f, "Debug"),
@@ -614,12 +642,21 @@ impl fmt::Debug for Instruction {
                 instance_function,
                 variadic,
                 generator,
+                arg_is_unpacked_tuple,
                 size,
             } => write!(
                 f,
-                "Function\tresult: {}\targs: {}\t\tcaptures: {}\tsize: {}\n\
-                     \t\t\tinstance: {}\tvariadic: {}\tgenerator: {}",
-                register, arg_count, capture_count, size, instance_function, variadic, generator,
+                "Function\tresult: {}\targs: {}\t\tcaptures: {}\tsize: {}
+                 \t\t\tinstance: {}\tgenerator: {}
+                 \t\t\tvariadic: {}\targ_is_unpacked_tuple: {}",
+                register,
+                arg_count,
+                capture_count,
+                size,
+                instance_function,
+                generator,
+                variadic,
+                arg_is_unpacked_tuple,
             ),
             Capture {
                 function,
@@ -630,8 +667,11 @@ impl fmt::Debug for Instruction {
                 "Capture\t\tfunction: {}\ttarget: {}\tsource: {}",
                 function, target, source
             ),
-            Negate { register, source } => {
-                write!(f, "Negate\t\tresult: {}\tsource: {}", register, source)
+            Negate { register, value } => {
+                write!(f, "Negate\t\tresult: {}\tsource: {}", register, value)
+            }
+            Not { register, value } => {
+                write!(f, "Not\t\tresult: {}\tsource: {}", register, value)
             }
             Add { register, lhs, rhs } => write!(
                 f,
@@ -699,15 +739,6 @@ impl fmt::Debug for Instruction {
                 register, offset, jump_condition
             ),
             JumpBack { offset } => write!(f, "JumpBack\toffset: {}", offset),
-            JumpBackIf {
-                register,
-                offset,
-                jump_condition,
-            } => write!(
-                f,
-                "JumpBackIf\tresult: {}\toffset: {}\tcondition: {}",
-                register, offset, jump_condition
-            ),
             Call {
                 result,
                 function,
@@ -718,16 +749,17 @@ impl fmt::Debug for Instruction {
                 "Call\t\tresult: {}\tfunction: {}\tframe base: {}\targs: {}",
                 result, function, frame_base, arg_count
             ),
-            CallChild {
+            CallInstance {
                 result,
                 function,
-                parent,
                 frame_base,
                 arg_count,
+                instance,
             } => write!(
                 f,
-                "CallChild\tresult: {}\tfunction: {}\tframe_base: {}\n\t\t\targs: {}\t\tparent: {}",
-                result, function, frame_base, arg_count, parent
+                "CallInstance\tresult: {}\tfunction: {}\tframe_base: {}
+                 \t\t\targs: {}\t\tinstance: {}",
+                result, function, frame_base, arg_count, instance
             ),
             Return { register } => write!(f, "Return\t\tresult: {}", register),
             Yield { register } => write!(f, "Yield\t\tresult: {}", register),
@@ -759,13 +791,13 @@ impl fmt::Debug for Instruction {
                 "IterNextQuiet\titerator: {}\tjump offset: {}",
                 iterator, jump_offset
             ),
-            ValueIndex {
+            TempIndex {
                 register,
                 value,
                 index,
             } => write!(
                 f,
-                "ValueIndex\tresult: {}\tvalue: {}\tindex: {}",
+                "TempIndex\tresult: {}\tvalue: {}\tindex: {}",
                 register, value, index
             ),
             SliceFrom {
@@ -853,6 +885,15 @@ impl fmt::Debug for Instruction {
                 "Access\t\tresult: {}\tvalue: {}\tkey: {}",
                 register, value, key
             ),
+            AccessString {
+                register,
+                value,
+                key,
+            } => write!(
+                f,
+                "AccessString\tresult: {}\tvalue: {}\tkey: {}",
+                register, value, key
+            ),
             TryStart {
                 arg_register,
                 catch_offset,
@@ -871,8 +912,15 @@ impl fmt::Debug for Instruction {
             CheckSize { register, size } => {
                 write!(f, "CheckSize\tregister: {}\tsize: {}", register, size)
             }
-            StringStart { register } => {
-                write!(f, "StringStart\tregister: {}", register)
+            StringStart {
+                register,
+                size_hint,
+            } => {
+                write!(
+                    f,
+                    "StringStart\tregister: {}\tsize hint: {}",
+                    register, size_hint
+                )
             }
             StringPush { register, value } => {
                 write!(f, "StringPush\tregister: {}\tvalue: {}", register, value)
@@ -887,12 +935,15 @@ impl fmt::Debug for Instruction {
 /// An iterator that converts bytecode into a series of [Instruction]s
 #[derive(Clone, Default)]
 pub struct InstructionReader {
-    pub chunk: Arc<Chunk>,
+    /// The chunk that the reader is reading from
+    pub chunk: Rc<Chunk>,
+    /// The reader's instruction pointer
     pub ip: usize,
 }
 
 impl InstructionReader {
-    pub fn new(chunk: Arc<Chunk>) -> Self {
+    /// Initializes a reader with the given chunk
+    pub fn new(chunk: Rc<Chunk>) -> Self {
         Self { chunk, ip: 0 }
     }
 }
@@ -905,6 +956,14 @@ impl Iterator for InstructionReader {
 
         macro_rules! get_u8 {
             () => {{
+                #[cfg(not(debug_assertions))]
+                {
+                    let byte = unsafe { self.chunk.bytes.get_unchecked(self.ip) };
+                    self.ip += 1;
+                    *byte
+                }
+
+                #[cfg(debug_assertions)]
                 match self.chunk.bytes.get(self.ip) {
                     Some(byte) => {
                         self.ip += 1;
@@ -921,15 +980,25 @@ impl Iterator for InstructionReader {
 
         macro_rules! get_u16 {
             () => {{
-                match self.chunk.bytes.get(self.ip..self.ip + 2) {
-                    Some(u16_bytes) => {
-                        self.ip += 2;
-                        u16::from_le_bytes(u16_bytes.try_into().unwrap())
-                    }
-                    None => {
-                        return Some(Error {
-                            message: format!("Expected 2 bytes at position {}", self.ip),
-                        });
+                #[cfg(not(debug_assertions))]
+                {
+                    let bytes = unsafe { self.chunk.bytes.get_unchecked(self.ip..self.ip + 2) };
+                    self.ip += 2;
+                    u16::from_le_bytes(bytes.try_into().unwrap())
+                }
+
+                #[cfg(debug_assertions)]
+                {
+                    match self.chunk.bytes.get(self.ip..self.ip + 2) {
+                        Some(u16_bytes) => {
+                            self.ip += 2;
+                            u16::from_le_bytes(u16_bytes.try_into().unwrap())
+                        }
+                        None => {
+                            return Some(Error {
+                                message: format!("Expected 2 bytes at position {}", self.ip),
+                            });
+                        }
                     }
                 }
             }};
@@ -937,15 +1006,81 @@ impl Iterator for InstructionReader {
 
         macro_rules! get_u32 {
             () => {{
-                match self.chunk.bytes.get(self.ip..self.ip + 4) {
-                    Some(u32_bytes) => {
-                        self.ip += 4;
-                        u32::from_le_bytes(u32_bytes.try_into().unwrap())
+                #[cfg(not(debug_assertions))]
+                {
+                    let bytes = unsafe { self.chunk.bytes.get_unchecked(self.ip..self.ip + 4) };
+                    self.ip += 4;
+                    u32::from_le_bytes(bytes.try_into().unwrap())
+                }
+
+                #[cfg(debug_assertions)]
+                {
+                    match self.chunk.bytes.get(self.ip..self.ip + 4) {
+                        Some(u32_bytes) => {
+                            self.ip += 4;
+                            u32::from_le_bytes(u32_bytes.try_into().unwrap())
+                        }
+                        None => {
+                            return Some(Error {
+                                message: format!("Expected 4 bytes at position {}", self.ip),
+                            });
+                        }
                     }
-                    None => {
-                        return Some(Error {
-                            message: format!("Expected 4 bytes at position {}", self.ip),
-                        });
+                }
+            }};
+        }
+
+        macro_rules! get_u16_constant {
+            () => {{
+                #[cfg(not(debug_assertions))]
+                {
+                    let bytes = unsafe { self.chunk.bytes.get_unchecked(self.ip..self.ip + 2) };
+                    self.ip += 2;
+                    let bytes: [u8; 2] = bytes.try_into().unwrap();
+                    ConstantIndex::from(bytes)
+                }
+
+                #[cfg(debug_assertions)]
+                {
+                    match self.chunk.bytes.get(self.ip..self.ip + 2) {
+                        Some(bytes) => {
+                            self.ip += 2;
+                            let bytes: [u8; 2] = bytes.try_into().unwrap();
+                            ConstantIndex::from(bytes)
+                        }
+                        None => {
+                            return Some(Error {
+                                message: format!("Expected 2 bytes at position {}", self.ip),
+                            });
+                        }
+                    }
+                }
+            }};
+        }
+
+        macro_rules! get_u24_constant {
+            () => {{
+                #[cfg(not(debug_assertions))]
+                {
+                    let bytes = unsafe { self.chunk.bytes.get_unchecked(self.ip..self.ip + 3) };
+                    self.ip += 3;
+                    let bytes: [u8; 3] = bytes.try_into().unwrap();
+                    ConstantIndex::from(bytes)
+                }
+
+                #[cfg(debug_assertions)]
+                {
+                    match self.chunk.bytes.get(self.ip..self.ip + 3) {
+                        Some(bytes) => {
+                            self.ip += 3;
+                            let bytes: [u8; 3] = bytes.try_into().unwrap();
+                            ConstantIndex::from(bytes)
+                        }
+                        None => {
+                            return Some(Error {
+                                message: format!("Expected 3 bytes at position {}", self.ip),
+                            });
+                        }
                     }
                 }
             }};
@@ -989,51 +1124,51 @@ impl Iterator for InstructionReader {
             }),
             Op::LoadFloat => Some(LoadFloat {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), 0, 0),
+                constant: ConstantIndex::from(get_u8!()),
             }),
             Op::LoadFloat16 => Some(LoadFloat {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), 0),
+                constant: get_u16_constant!(),
             }),
             Op::LoadFloat24 => Some(LoadFloat {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), get_u8!()),
+                constant: get_u24_constant!(),
             }),
             Op::LoadInt => Some(LoadInt {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), 0, 0),
+                constant: ConstantIndex::from(get_u8!()),
             }),
             Op::LoadInt16 => Some(LoadInt {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), 0),
+                constant: get_u16_constant!(),
             }),
             Op::LoadInt24 => Some(LoadInt {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), get_u8!()),
+                constant: get_u24_constant!(),
             }),
             Op::LoadString => Some(LoadString {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), 0, 0),
+                constant: ConstantIndex::from(get_u8!()),
             }),
             Op::LoadString16 => Some(LoadString {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), 0),
+                constant: get_u16_constant!(),
             }),
             Op::LoadString24 => Some(LoadString {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), get_u8!()),
+                constant: get_u24_constant!(),
             }),
             Op::LoadNonLocal => Some(LoadNonLocal {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), 0, 0),
+                constant: ConstantIndex::from(get_u8!()),
             }),
             Op::LoadNonLocal16 => Some(LoadNonLocal {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), 0),
+                constant: get_u16_constant!(),
             }),
             Op::LoadNonLocal24 => Some(LoadNonLocal {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), get_u8!()),
+                constant: get_u24_constant!(),
             }),
             Op::ValueExport => Some(ValueExport {
                 name: get_u8!(),
@@ -1057,13 +1192,13 @@ impl Iterator for InstructionReader {
             }),
             Op::MakeNum2 => Some(MakeNum2 {
                 register: get_u8!(),
-                count: get_u8!(),
                 element_register: get_u8!(),
+                count: get_u8!(),
             }),
             Op::MakeNum4 => Some(MakeNum4 {
                 register: get_u8!(),
-                count: get_u8!(),
                 element_register: get_u8!(),
+                count: get_u8!(),
             }),
             Op::SequenceStart => Some(SequenceStart {
                 register: get_u8!(),
@@ -1142,6 +1277,7 @@ impl Iterator for InstructionReader {
                     instance_function: flags.instance_function,
                     variadic: flags.variadic,
                     generator: flags.generator,
+                    arg_is_unpacked_tuple: flags.arg_is_unpacked_tuple,
                     size,
                 })
             }
@@ -1152,7 +1288,11 @@ impl Iterator for InstructionReader {
             }),
             Op::Negate => Some(Negate {
                 register: get_u8!(),
-                source: get_u8!(),
+                value: get_u8!(),
+            }),
+            Op::Not => Some(Not {
+                register: get_u8!(),
+                value: get_u8!(),
             }),
             Op::Add => Some(Add {
                 register: get_u8!(),
@@ -1225,23 +1365,18 @@ impl Iterator for InstructionReader {
             Op::JumpBack => Some(JumpBack {
                 offset: get_u16!() as usize,
             }),
-            Op::JumpBackFalse => Some(JumpBackIf {
-                register: get_u8!(),
-                offset: get_u16!() as usize,
-                jump_condition: false,
-            }),
             Op::Call => Some(Call {
                 result: get_u8!(),
                 function: get_u8!(),
                 frame_base: get_u8!(),
                 arg_count: get_u8!(),
             }),
-            Op::CallChild => Some(CallChild {
+            Op::CallInstance => Some(CallInstance {
                 result: get_u8!(),
                 function: get_u8!(),
                 frame_base: get_u8!(),
                 arg_count: get_u8!(),
-                parent: get_u8!(),
+                instance: get_u8!(),
             }),
             Op::Return => Some(Return {
                 register: get_u8!(),
@@ -1270,7 +1405,7 @@ impl Iterator for InstructionReader {
                 iterator: get_u8!(),
                 jump_offset: get_u16!() as usize,
             }),
-            Op::ValueIndex => Some(ValueIndex {
+            Op::TempIndex => Some(TempIndex {
                 register: get_u8!(),
                 value: get_u8!(),
                 index: get_u8!() as i8,
@@ -1378,6 +1513,21 @@ impl Iterator for InstructionReader {
                 }
             }
             Op::Access => Some(Access {
+                register: get_u8!(),
+                value: get_u8!(),
+                key: ConstantIndex::from(get_u8!()),
+            }),
+            Op::Access16 => Some(Access {
+                register: get_u8!(),
+                value: get_u8!(),
+                key: get_u16_constant!(),
+            }),
+            Op::Access24 => Some(Access {
+                register: get_u8!(),
+                value: get_u8!(),
+                key: get_u24_constant!(),
+            }),
+            Op::AccessString => Some(AccessString {
                 register: get_u8!(),
                 value: get_u8!(),
                 key: get_u8!(),
@@ -1389,7 +1539,7 @@ impl Iterator for InstructionReader {
             Op::TryEnd => Some(TryEnd),
             Op::Debug => Some(Debug {
                 register: get_u8!(),
-                constant: ConstantIndex(get_u8!(), get_u8!(), get_u8!()),
+                constant: get_u24_constant!(),
             }),
             Op::CheckType => {
                 let register = get_u8!();
@@ -1406,6 +1556,11 @@ impl Iterator for InstructionReader {
             }),
             Op::StringStart => Some(StringStart {
                 register: get_u8!(),
+                size_hint: get_u8!() as usize,
+            }),
+            Op::StringStart32 => Some(StringStart {
+                register: get_u8!(),
+                size_hint: get_u32!() as usize,
             }),
             Op::StringPush => Some(StringPush {
                 register: get_u8!(),
